/*
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */

use std::ffi::OsStr;
use std::fs::read_to_string;
use std::io::ErrorKind;
use std::path::Path;

use mozjs::rust::JSEngine;
use sourcemap::SourceMap;

use ion::format::Config as FormatConfig;
use ion::format::format_value;
use modules::Modules;
use runtime::{Runtime, RuntimeBuilder};
use runtime::cache::locate_in_cache;
use runtime::config::Config;
use runtime::modules::Module;
use runtime::script::Script;

pub fn eval_inline(rt: &Runtime, source: &str) {
	let result = Script::compile_and_evaluate(rt.cx(), &Path::new("inline.js"), source);

	match result {
		Ok(v) => println!("{}", format_value(rt.cx(), FormatConfig::default().quoted(true), v)),
		Err(report) => eprintln!("{}", report),
	}
	if !rt.run_event_loop() {
		eprintln!("Unknown error occurred while executing microtask.");
	}
}

pub fn eval_script(path: &Path) {
	let engine = JSEngine::init().unwrap();
	let rt = RuntimeBuilder::<Modules>::new()
		.macrotask_queue()
		.microtask_queue()
		.standard_modules()
		.build(engine.handle());

	if let Some((script, _)) = read_script(path) {
		let (script, sourcemap) = cache(path, script);
		let result = Script::compile_and_evaluate(rt.cx(), path, &script);

		match result {
			Ok(v) => println!("{}", format_value(rt.cx(), FormatConfig::default().quoted(true), v)),
			Err(mut report) => {
				if let Some(sourcemap) = sourcemap {
					report.transform_with_sourcemap(&sourcemap);
				}
				eprintln!("{}", report);
			}
		}
		if !rt.run_event_loop() {
			eprintln!("Unknown error occurred while executing microtask.");
		}
	}
}

pub fn eval_module(path: &Path) {
	let engine = JSEngine::init().unwrap();
	let rt = RuntimeBuilder::<Modules>::new()
		.macrotask_queue()
		.microtask_queue()
		.modules()
		.standard_modules()
		.build(engine.handle());

	if let Some((script, filename)) = read_script(path) {
		let (script, sourcemap) = cache(path, script);
		let result = Module::compile(rt.cx(), &filename, Some(path), &script);

		if let Err(mut error) = result {
			if let Some(sourcemap) = sourcemap {
				error.inner_mut().transform_with_sourcemap(&sourcemap);
			}
			eprintln!("{}", error);
		}
		if !rt.run_event_loop() {
			eprintln!("Unknown error occurred while executing microtask.");
		}
	}
}

fn read_script(path: &Path) -> Option<(String, String)> {
	match read_to_string(path) {
		Ok(script) => {
			let filename = String::from(path.file_name().unwrap().to_str().unwrap());
			Some((script, filename))
		}
		Err(error) => {
			eprintln!("Failed to read file: {}", path.display());
			match error.kind() {
				ErrorKind::NotFound => eprintln!("(File was not found)"),
				ErrorKind::PermissionDenied => eprintln!("Current User lacks permissions to read the file)"),
				_ => eprintln!("{:?}", error),
			}
			None
		}
	}
}

fn cache(path: &Path, script: String) -> (String, Option<SourceMap>) {
	let is_typescript = Config::global().typescript && path.extension() == Some(OsStr::new("ts"));
	is_typescript
		.then(|| locate_in_cache(path, &script))
		.flatten()
		.map(|(s, sm)| (s, Some(sm)))
		.unwrap_or_else(|| (script, None))
}
